// Solace -- Sol Anachronistic Computer Emulation
// A Win32 emulator for the Sol-20 computer.
//
// Copyright (c) Jim Battle, 2001
//
// This file contains structures and function prototypes for
// manipulating virtual tapes for Solace, the Sol emulator.

#ifndef _VNSDISK_H_
#define _VNSDISK_H_

// some disk properties
#define SVD_HDR_SIZE      (4096)	// # bytes in svd header
#define SVD_LABEL_SIZE    (3*1024)	// # bytes in svd label
#define SVD_MINTRACKS     (10)
#define SVD_MAXTRACKS     (96)
#define SVD_FILENAME_SIZE (300)	// maximum fully specified filename size


// DISK FORMAT:
//
// Each disk consists of a header and a fixed number of blocks.
// The header occupies the first 4 KB of the file and is stored
// as CR/LF-terminated ASCII strings, with any remainder of the 4 KB
// padded with 0x00 bytes.
//
// This header block has this fixed sequence:
//
//   struct {
//     char format[64];     "SVD:Solace Virtual Disk, Northstar\00\00..."
//     uint16 version;      [=1]
//     uint16 density;      [=1|2]
//     uint16 writeprotect; [=0|1]
//     uint16 sides;        [=1|2]
//     uint16 tracks;       [=35-96]
//     uint16 sectors;      [=10]
//     unsigned char pad[1024 - 64 - 6*2];
//     char label[3*1024];
//   }
//
// The format string is a null-terminated string that identifies
// what the rest of the header block might contain.  Any unused
// portion of the string should be set to 0x00.
//
// All the uint16 fields are stored little-endian, meaning that the
// least-significant byte of the 16b quantity is stored earlier in
// the file than the most-significant byte.
//
// Density indicates if the entire disk is single or double density.
// This field is used to minimize the file size if all sectors are
// known to be single-density.  It is OK to write nothing but single-
// density sectors to a double-density disk, or any mix, of course.
//
// If writeprotect is 1, the disk shouldn't be written to; if it is 0,
// it is OK for the driver to write data to the disk.
//
// Sides indicates the number of sides that are writable on the disk.
// Single sided virtual disks take up less space on the real disk.
//
// Tracks is the number of tracks per side; it is 35 for the Northstar
// controller, but room has been left in the header for generality.
//
// Sectors is the number of sectors per track; it is fixed at 10 for
// the Northstar controller, but room has been left in the header
// for generality.
//
// Pad should be filled with 0x00 bytes to round out the first 1KB
// of the header block.
//
// The label lines are a single null-terminated string, although
// there can be any number of lines, which are CR/LF terminated.
//
// The rest of the file is accessed randomly as fixed size blocks.
// These blocks are saved as a raw binary format.
//
// The number of blocks is (# sides)*(# tracks)*(10).
//
// The size of each block depends on the density.  Nominally there
// are 256 bytes/sector in single density, and 512 bytes/sector in
// double density.  However, the actual number (unformatted) is
// dependent only on the rotational speed and the bit transfer rate.
// At the nominal 300 RPM, each rotation takes 200 ms.  At a transfer
// rate of 250 Kbps, this works out to 6250 bytes/track double density,
// and 3125 bytes/track single density.  In real life not all these
// bytes can be used, as the disk speed might be higher or lower,
// and each sector requires preamble and sync bytes.
//      single density (density==1): 256 +  56 = 312 bytes/block
//      double density (density==2): 512 + 112 = 624 bytes/block
//
// Although the standard Northstar format only uses a few of these
// "extra" bytes, for preamble and checksum, in theory they can be
// used and Solace should work for any such scheme too.
//
// The file that contains the virtual disk image adds yet another
// four bytes of overhead per sector.  Only one bit of this is
// currently used, which is for indicating if the sector is
// recorded in single or double density.
//
// Having a block size that isn't a power of two doesn't match very
// nicely to the actual disk blocking factor, but it is still tons
// faster than the floppy it is emulating, so no harm done.

// return status used by various routines
#define SVD_OK           0	// everything is OK
#define SVD_FILEEXISTS  -1	// file already exists on disk
#define SVD_NOFILE      -2	// requested file doesn't exist
#define SVD_BADFILE     -3	// bad file format
#define SVD_OPENERROR   -4	// couldn't open file
#define SVD_ACCESSERROR -5	// couldn't read or write to file
#define SVD_BADFORMAT   -6	// the format of this disk is wrong

// return a descriptive string to decipher numeric error code
char *vdisk_errstring(int err);


// ===================================================================
//               Logical Layer Access Functions
//
// These functions are used to access the virtual disk independently
// of the controller hardware.
// ===================================================================

// given a filename, get information about a disk.
// returns SVD_OK if it worked, something else if it didn't.
int svd_read_disk_info( char *filename,
			int  *filetype,
			int  *sides,   int *tracks,       int  *sectors,
			int  *density, int *writeprotect, char *label,
			int  *maxsectorsize);

// write a disk header to the specified file.
// returns SVD_OK if it worked, something else if it didn't.
int svd_write_disk_info( char *filename,
			 int   sides, int tracks, int sectors,
			 int   density, int writeprotect, char *label);

// create a blank virtual disk of the specified type
// return 0 if OK, otherwise some error indication.
// does not check if file already exists.
int svd_format(char *filename, int sides, int tracks, int sectors,
	       int density, int writeprotect, char *label, char fillchar);

// install boot tracks on an already formatted disk.
int svd_make_bootable(char *filename, byte *image, int tracks);

// debugging aid -- dump contents of virtual disk file
int svd_dump_disk(char *diskfilename, char *dumpfilename);


// ===================================================================
//                       Emulation Layer
//
// These functions are used to manipulate the emulated hardware --
// the controller and virtual disk drives.
// ===================================================================

// called once at power up
void svd_emu_init(void);

// must be called on reset
void svd_board_reset(void);

// install the card in the system
int svd_install_driver(int base_address);

// this gives the driver a chance to clean up, such as flushing
// any buffers to disk.  returning 0 means OK, 1 means request denied.
int svd_request_uninstall_driver(void);

// pull the card from the system
int svd_uninstall_driver(void);

// put a disk into a drive
int svd_insert_disk(int drive, char *filename);

// remove a disk from a drive
void svd_remove_disk(int drive);


// virtual disk controller properties
// later add more props for debug mode control options
enum { VDCPROP_INSTALLED = 1,	// controller installed or not
       VDCPROP_BASEADDR  = 2,	// where in S-100 space
       VDCPROP_DRIVES    = 3,	// number of disk drives
       VDCPROP_DEBUGMODE = 4	// complain when SW driver errs
     };
// these return 0 on success, non-zero on failure
enum {	VDCPROP_OK      = 0,	// success
	VDCPROP_BADPROP = 1,	// no such property
	VDCPROP_CANTSET = 2,	// not allowed to set this property
	VDCPROP_NOTNOW  = 3	// this property can be set, just not now
     };

int GetDiskControllerProp(int prop, int *value);
int SetDiskControllerProp(int prop, int  value);


// this interface allows querying and in some cases setting
// the state of a given disk or disk drive.
// properties #1 to #7 are read-only
enum { VDPROP_EMPTY    = 1,	// read-only; drive state
       VDPROP_SELECTED = 2,	// read-only; drive state
       VDPROP_SIDES    = 3,	// read-only; disk state
       VDPROP_TRACKS   = 4,	// read-only; disk state
       VDPROP_SECTORS  = 5,	// read-only; disk state
       VDPROP_DENSITY  = 6,	// read-only; disk state
       VDPROP_READONLY = 7,	// disk state (write protect tab)
       VDPROP_LABEL    = 8,	// disk state
       VDPROP_FILENAME = 9 };

enum { VDFORMAT_NORTHSTAR=0, VDFORMAT_HELIOS=10 };

enum {	VDPROP_OK       = 0,	// success
	VDPROP_BADPROP  = 1,	// no such property
	VDPROP_CANTSET  = 2,	// not allowed to set this property
	VDPROP_BADDRIVE = 3,	// not a legal disk drive
	VDPROP_NODISK   = 4	// no disk in drive
     };

// these return 0 on success, non-zero on failure
int GetDiskProp(int unit, int prop, int *value);
int SetDiskProp(int unit, int prop, int  value);


// return 1 if the specified virtual disk file is already in use,
// return 0 otherwise.
int DiskInUse(char *filename);


#endif // ifdef _VNSDISK_H_

